# -*- coding: utf-8 -*-

import pytest


@pytest.mark.usefixtures("as_connection")
class TestGetKeyDigest(object):
    def test_get_key_digest_with_integer_key(self):
        """
        Invoke get_key_digest() with integer key
        """

        digest = self.as_connection.get_key_digest("test", "demo", 1)

        assert isinstance(digest, bytearray)

    def test_get_key_digest_with_string_key(self):
        """
        Invoke get_key_digest() with string key
        """

        digest = self.as_connection.get_key_digest("test", "demo", "get_digest_key")

        assert isinstance(digest, bytearray)

    def test_get_key_digest_with_byte_key(self):
        """
        Test calling the method with a bytearray primary key
        """
        key = bytearray("a" * 5, "utf-8")
        digest = self.as_connection.get_key_digest("test", "demo", key)

        assert isinstance(digest, bytearray)

    @pytest.mark.xfail()
    def test_get_key_digest_with_none_set(self):
        """
        Test the value of calling calc digest with None for a set
        """
        digest = self.as_connection.get_key_digest("test", None, 1)

        assert isinstance(digest, bytearray)

    # **Beginning of tests for invalid usage**

    def test_get_key_digest_with_no_parameter(self):
        """
        Invoke get_key_digest() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.get_key_digest()

        assert "argument 'ns' (pos 1)" in str(typeError.value)

    def test_get_key_digest_with_only_ns(self):
        """
        Invoke get_key_digest() with a ns and no other parameters
        """
        with pytest.raises(TypeError):
            self.as_connection.get_key_digest("test")

    def test_get_key_digest_with_only_ns_and_set(self):
        """
        Invoke get_key_digest() with a ns and no other parameters
        """
        with pytest.raises(TypeError):
            self.as_connection.get_key_digest("test", "set")

    @pytest.mark.parametrize("ns", [1, None, False, {}, [], ()])
    def test_get_key_digest_with_invalid_ns_type(self, ns):
        """
        Invoke get_key_digest() with integer ns
        """

        with pytest.raises(TypeError):
            self.as_connection.get_key_digest(ns, "demo", "get_digest_key")

    @pytest.mark.parametrize("set_val", [1, None, False, {}, [], ()])
    def test_get_key_digest_with_integer_set(self, set_val):
        """
        Invoke get_key_digest() with integer set
        """

        with pytest.raises(TypeError):
            self.as_connection.get_key_digest("test", set_val, "get_digest_key")

    @pytest.mark.parametrize(
        "namespace, set, primary_key",
        [("test", "demo", [1, 2, 3]), ("test", "demo", {"a": 1, "b": 2}), ("test", "demo", None)],
        ids=("list key", "map key", "None key"),
    )
    def test_get_key_digest_with_invalid_key_type(self, namespace, set, primary_key):
        with pytest.raises(TypeError):
            self.as_connection.get_key_digest(namespace, set, primary_key)



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################


from __future__ import print_function
import aerospike
import pprint
import sys

from optparse import OptionParser

##########################################################################
# Options Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")
optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Application
##########################################################################

exitCode = 0

try:
    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    config = {'hosts': [(options.host, options.port)]}
    client = aerospike.client(config).connect(
        options.username, options.password)
    # ----------------------------------------------------------------------------
    # Perform Operations
    # ----------------------------------------------------------------------------

    try:
        pp = pprint.PrettyPrinter(indent=2)
        client.put(('test', 'cats', 'mr. peppy'), {'breed': 'persian'},
                   policy={'exists': aerospike.POLICY_EXISTS_CREATE_OR_REPLACE,
                           'key': aerospike.POLICY_KEY_DIGEST},
                   meta={'ttl': 120})
        (key, meta, bins) = client.get(('test', 'cats', 'mr. peppy'))
        print("Before:", bins)
        client.increment(
            key, 'lives', -1, {'gen': 2, 'ttl': 1000}, policy={'total_timeout': 1500})
        (key, meta, bins) = client.get(key)
        print("After:", bins)
        # the key we got back when we fetched the record with get() is useable
        # as-is because it contains the record's digest
        client.increment(key, 'lives', -1)
        (key, meta, bins) = client.get(key)
        # kitty lost a life, unfortunately
        print("Poor Kitty:", bins)
        client.put(key, {'owner': 'Fry'})
        client.prepend(key, 'owner', 'Philip J. ')
        client.append(key, 'owner', ' Esq.')
        # kitty loses another life, gains a color, all as part of a record
        # multi-op
        ops = [{'bin': 'color', 'op': aerospike.OPERATOR_WRITE, 'val': 'smoke'},
               {'bin': 'lives', 'op': aerospike.OPERATOR_INCR, 'val': -1},
               {'bin': 'ailments', 'op': aerospike.OPERATOR_READ},
               {'bin': 'lives', 'op': aerospike.OPERATOR_READ}]
        (key, meta, bins) = client.operate(key, ops)
        print("After calling operate(), kitty is down to",
              bins['lives'], "lives")
        pp.pprint(bins)

        # display the record as it is after all the operations
        (key, meta, bins) = client.get(key)
        print("\nRecord\n======\nKey\n---")
        pp.pprint(key)
        print("Meta\n----")
        pp.pprint(meta)
        print("Bins\n----")
        pp.pprint(bins)

    except Exception as eargs:
        print("error: {0}".format(eargs), file=sys.stderr)
        exitCode = 2

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3

##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)



import time

client.put(("test", "demo", "key1"), {"bin": 4})

time.sleep(1)
# Take threshold time
current_time = time.time()
time.sleep(1)

client.put(("test", "demo", "key2"), {"bin": 5})

threshold_ns = int(current_time * 10**9)
# Remove all items in set `demo` created before threshold time
# Record using key1 should be removed
client.truncate('test', 'demo', threshold_ns)

# Remove all items in namespace
# client.truncate('test', None, 0)



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e

import aerospike
from .test_base_class import TestBaseClass

Server61 = True

if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
    if pytest.__version__ < "3.0.0":
        Server61 = False
    else:
        Server61 = False


def get_list_result_from_operation(client, key, operation, bin):
    _, _, result_bins = client.operate(key, [operation])
    return result_bins[bin]


class TestNewListOperations(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        # INDEXES    0, 1, 2, 3, 4, 05
        # RINDEX     5, 4, 3, 2, 1, 0
        # RANK       2, 1, 0, 3, 4, 5
        # RRANK     -4,-5,-6,-3,-2,-1
        self.test_list = [7, 6, 5, 8, 9, 10]

        self.test_key = "test", "demo", "new_list_op"
        self.test_bin = "list"
        self.as_connection.put(self.test_key, {self.test_bin: self.test_list})
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    @pytest.mark.parametrize("index, expected", ((2, 5), (-2, 9)))
    def test_get_by_index(self, index, expected):
        """
        Without a return type this should return the value
        """
        operation = {
            "op": aerospike.OP_LIST_GET_BY_INDEX,
            "bin": "list",
            "index": index,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == expected

    @pytest.mark.parametrize(
        "return_type, expected",
        (
            (aerospike.LIST_RETURN_VALUE, 5),
            (aerospike.LIST_RETURN_INDEX, 2),
            (aerospike.LIST_RETURN_REVERSE_INDEX, 3),
            (aerospike.LIST_RETURN_RANK, 0),
            (aerospike.LIST_RETURN_REVERSE_RANK, 5),
        ),
    )
    def test_list_get_by_index_return_types(self, return_type, expected):
        """
        Without a return type this should return the value
        """
        operation = {"op": aerospike.OP_LIST_GET_BY_INDEX, "bin": "list", "index": 2, "return_type": return_type}
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == expected

    @pytest.mark.parametrize("index, count", ((0, 3), (-2, 1), (0, 100)))
    def test_get_by_index_range_both_present(self, index, count):
        expected = self.test_list[index : index + count]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_INDEX_RANGE,
            "bin": self.test_bin,
            "index": index,
            "count": count,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == expected

    def test_get_by_index_range_no_count(self):
        operation = {
            "op": aerospike.OP_LIST_GET_BY_INDEX_RANGE,
            "bin": self.test_bin,
            "index": 2,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == self.test_list[2:]

    def test_get_by_index_range_inverted(self):
        start = 0
        count = 3
        expected = self.test_list[start + count :]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_INDEX_RANGE,
            "bin": self.test_bin,
            "index": start,
            "count": count,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == expected

    @pytest.mark.parametrize("rank, expected", ((0, 5), (-1, 10)))
    def test_get_by_rank(self, rank, expected):
        """
        Without a return type this should return the value
        """
        operation = {
            "op": aerospike.OP_LIST_GET_BY_RANK,
            "bin": "list",
            "rank": rank,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == expected

    @pytest.mark.parametrize(
        "return_type, expected",
        (
            (aerospike.LIST_RETURN_VALUE, 5),
            (aerospike.LIST_RETURN_INDEX, 2),
            (aerospike.LIST_RETURN_REVERSE_INDEX, 3),
            (aerospike.LIST_RETURN_RANK, 0),
            (aerospike.LIST_RETURN_REVERSE_RANK, 5),
        ),
    )
    def test_list_get_by_rank_return_types(self, return_type, expected):
        """
        Without a return type this should return the value
        """
        operation = {"op": aerospike.OP_LIST_GET_BY_RANK, "bin": "list", "rank": 0, "return_type": return_type}
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == expected

    @pytest.mark.parametrize("rank, count", ((0, 3), (-2, 1), (0, 100)))
    def test_get_by_rank_range_both_present(self, rank, count):
        expected = sorted(self.test_list)[rank : rank + count]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_RANK_RANGE,
            "bin": self.test_bin,
            "rank": rank,
            "count": count,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert set(result) == set(expected)

    def test_get_by_rank_range_no_count(self):
        operation = {
            "op": aerospike.OP_LIST_GET_BY_RANK_RANGE,
            "bin": self.test_bin,
            "rank": 2,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == sorted(self.test_list)[2:]

    def test_get_by_rank_range_inverted(self):
        rank_start = 0
        rank_count = 3
        # All of the values except for those in the specified rank range
        expected = sorted(self.test_list)[rank_start + rank_count :]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_RANK_RANGE,
            "bin": self.test_bin,
            "rank": rank_start,
            "count": rank_count,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert set(result) == set(expected)

    def test_get_by_value_no_duplicates(self):
        """
        7 is in the 0th position, so we expect [0] as the result
        """
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE,
            "bin": self.test_bin,
            "val": 7,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == [0]

    def test_get_by_value_no_duplicates_inverted(self):
        """
        7 is in the 0th position, so we expect [0] as the result
        """
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE,
            "bin": self.test_bin,
            "val": 7,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        # Every value except for 7
        assert result == [6, 5, 8, 9, 10]

    def test_get_by_value_with_duplicates(self):
        """
        Add a list [0, 1, 0, 2, 0] with 3 0's
        We expect [0, 2, 4] as the results
        """
        dup_list = [0, 1, 0, 2, 0]
        dup_key = "test", "list", "dup"
        self.keys.append(dup_key)

        self.as_connection.put(dup_key, {self.test_bin: dup_list})

        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE,
            "bin": self.test_bin,
            "val": 0,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, dup_key, operation, self.test_bin)
        assert result == [0, 2, 4]

    def test_get_by_value_list(self):
        values = [7, 5, 9]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_LIST,
            "bin": self.test_bin,
            "value_list": values,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == [0, 2, 4]

    def test_get_by_value_list_inverted(self):
        values = [7, 5, 9]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_LIST,
            "bin": self.test_bin,
            "value_list": values,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert set(result) == set([6, 8, 10])

    def test_get_by_value_range(self):
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_begin": 5,
            "value_end": 8,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert len(result) == 3 and set(result) == set([0, 1, 2])

    def test_get_by_value_range_inverted(self):
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_begin": 6,
            "value_end": 8,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert len(result) == 4 and set(result) == set([5, 8, 9, 10])

    def test_get_exists_by_index_range_no_count(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operation = {
            "op": aerospike.OP_LIST_GET_BY_INDEX_RANGE,
            "bin": self.test_bin,
            "index": 2,
            "return_type": aerospike.LIST_RETURN_EXISTS,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result is True

    def test_get_exists_by_index_range_inverted(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        start = 0
        count = 3
        expected = True
        operation = {
            "op": aerospike.OP_LIST_GET_BY_INDEX_RANGE,
            "bin": self.test_bin,
            "index": start,
            "count": count,
            "return_type": aerospike.LIST_RETURN_EXISTS,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == expected

    def test_get_exists_by_rank_range_no_count(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operation = {
            "op": aerospike.OP_LIST_GET_BY_RANK_RANGE,
            "bin": self.test_bin,
            "rank": 2,
            "return_type": aerospike.LIST_RETURN_EXISTS,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result is True

    def test_get_exists_by_value_no_duplicates(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        """
        7 is in the 0th position, so we expect it exists
        """
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE,
            "bin": self.test_bin,
            "val": 7,
            "return_type": aerospike.LIST_RETURN_EXISTS,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result is True

    def test_get_exists_by_value_with_duplicates(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        """
        Add a list [0, 1, 0, 2, 0] with 3 0's
        We expect True as the results
        """
        dup_list = [0, 1, 0, 2, 0]
        dup_key = "test", "list", "dup"
        self.keys.append(dup_key)

        self.as_connection.put(dup_key, {self.test_bin: dup_list})

        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE,
            "bin": self.test_bin,
            "val": 0,
            "return_type": aerospike.LIST_RETURN_EXISTS,
        }
        result = get_list_result_from_operation(self.as_connection, dup_key, operation, self.test_bin)
        assert result is True

    def test_get_exists_by_value_list(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        values = [7, 5, 9]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_LIST,
            "bin": self.test_bin,
            "value_list": values,
            "return_type": aerospike.LIST_RETURN_EXISTS,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result is True

    def test_get_exists_by_value_list_inverted(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        values = [7, 5, 9]
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_LIST,
            "bin": self.test_bin,
            "value_list": values,
            "return_type": aerospike.LIST_RETURN_EXISTS,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result is True

    def test_get_exists_by_value_range(self):
        if not Server61:
            pytest.skip("It only applies to >= 6.1 enterprise edition")
        operation = {
            "op": aerospike.OP_LIST_GET_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_begin": 5,
            "value_end": 8,
            "return_type": aerospike.LIST_RETURN_EXISTS,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result is True

    #  REMOVE Family of operations
    def test_remove_by_index(self):
        """
        Remove the 3rd item, a 5
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_INDEX,
            "bin": "list",
            "index": 2,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == 5
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[:2] + self.test_list[3:]

    def test_remove_by_index_range(self):
        """
        Remove the 3rd item, a 5
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_INDEX_RANGE,
            "bin": "list",
            "index": 2,
            "count": 2,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == [5, 8]
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[:2] + self.test_list[4:]

    def test_remove_by_index_range_inverted(self):
        """
        Remove the 3rd item, a 5
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_INDEX_RANGE,
            "bin": "list",
            "index": 2,
            "count": 2,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert set(result) == set([7, 6, 9, 10])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [5, 8]

    def test_remove_by_rank(self):
        """
        Remove the 3rd smallest item, a 7 at index 0
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_RANK,
            "bin": "list",
            "rank": 2,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == 7
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[1:]

    def test_remove_by_rank_range(self):
        """
        Remove the 3rd smallest item, a 7 at index 0
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_RANK_RANGE,
            "bin": "list",
            "rank": 0,
            "count": 3,
            "return_type": aerospike.LIST_RETURN_VALUE,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == [7, 6, 5]
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[3:]

    def test_remove_by_rank_range_inverted(self):
        """
        Remove the 3rd smallest item, a 7 at index 0
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_RANK_RANGE,
            "bin": "list",
            "rank": 0,
            "count": 3,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert set(result) == set([8, 9, 10])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[0:3]

    def test_remove_by_value_no_duplicates(self):
        """
        7 is in the 0th position, so we expect [0] as the result
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE,
            "bin": self.test_bin,
            "val": 7,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == [0]
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[1:]

    def test_remove_by_value_inverted(self):
        """
        7 is in the 0th position, so we expect [0] as the result
        """
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE,
            "bin": self.test_bin,
            "val": 7,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == [6, 5, 8, 9, 10]
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == self.test_list[0:1]

    def test_remove_by_value_with_duplicates(self):
        """
        Add a list [0, 1, 0, 2, 0] with 3 0's
        We expect [0, 2, 4] as the results
        """
        dup_list = [0, 1, 0, 2, 0]
        dup_key = "test", "list", "dup"
        self.keys.append(dup_key)

        self.as_connection.put(dup_key, {self.test_bin: dup_list})

        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE,
            "bin": self.test_bin,
            "val": 0,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, dup_key, operation, self.test_bin)
        assert result == [0, 2, 4]

        _, _, bins = self.as_connection.get(dup_key)
        assert bins[self.test_bin] == [1, 2]

    def test_remove_by_value_list(self):
        values = [7, 5, 9]
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE_LIST,
            "bin": self.test_bin,
            "value_list": values,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == [0, 2, 4]

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [6, 8, 10]

    def test_remove_by_value_list_inverted(self):
        values = [7, 5, 9]
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE_LIST,
            "bin": self.test_bin,
            "value_list": values,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert set(result) == set([6, 8, 10])

        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [7, 5, 9]

    def test_remove_by_value_range(self):
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_begin": 5,
            "value_end": 8,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert len(result) == 3 and set(result) == set([0, 1, 2])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [8, 9, 10]

    def test_remove_by_value_range_inverted(self):
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_begin": 6,
            "value_end": 8,
            "return_type": aerospike.LIST_RETURN_VALUE,
            "inverted": True,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert len(result) == 4 and set(result) == set([5, 8, 9, 10])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [7, 6]

    def test_remove_by_value_range_no_begin(self):
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_end": 8,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert len(result) == 3 and set(result) == set([0, 1, 2])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [8, 9, 10]

    def test_remove_by_value_range_no_end(self):
        operation = {
            "op": aerospike.OP_LIST_REMOVE_BY_VALUE_RANGE,
            "bin": self.test_bin,
            "value_begin": 7,
            "return_type": aerospike.LIST_RETURN_INDEX,
        }
        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert len(result) == 4 and set(result) == set([0, 3, 4, 5])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == [6, 5]

    def test_list_set_order(self):
        operation = {"op": aerospike.OP_LIST_SET_ORDER, "list_order": aerospike.LIST_ORDERED, "bin": self.test_bin}
        self.as_connection.operate(self.test_key, [operation])
        _, _, bins = self.as_connection.get(self.test_key)
        assert bins[self.test_bin] == sorted(self.test_list)

    def test_list_sort(self):
        unsorted_dups = [2, 5, 2, 5]
        sort_key = "test", "demo", "dup_list"
        self.keys.append(sort_key)
        self.as_connection.put(sort_key, {self.test_bin: unsorted_dups})

        operation = {
            "op": aerospike.OP_LIST_SORT,
            "sort_flags": aerospike.LIST_SORT_DROP_DUPLICATES,
            "bin": self.test_bin,
        }

        self.as_connection.operate(sort_key, [operation])
        _, _, bins = self.as_connection.get(sort_key)
        assert bins[self.test_bin] == [2, 5]



# -*- coding: utf-8 -*-
"""
Helper functions for dealing with the delayed nature of adding and removing UDFs
"""
import time


def wait_for_udf_removal(connection, udf_name, sleep_time=0.25, attempts=8):
    """
    This should be called after a udf_removal to wait for the removal to take place
    If this takes more than sleep_time * attempts seconds, it will quit and return False
    Sets the udf_removed flag on the class, so the removal can be skipped on the teardown method
    """
    iterations = 0
    present = True

    # While the UDF is still in the UDF List, loop
    while present:

        # Break when this has run beyond the attempts threshold
        if iterations > attempts:
            return False

        udf_list = connection.udf_list({"timeout": 100})

        found = False
        # Check if the UDF is in the returned list
        for udf in udf_list:
            if udf["name"] == udf_name:
                found = True
                break

        present = found

        # If it is still present, increment our iterations and pause
        if present:
            iterations += 1
            time.sleep(sleep_time)

    return True


def wait_for_udf_to_exist(connection, udf_name, sleep_time=0.25, attempts=8):
    """
    This should be called after a udf_put to wait for the udf to be stored on the server.
    """
    iterations = 0
    exists = False

    # Loop while the UDF is not found on the server
    while not exists:

        # if we have tried too many times, exit
        if iterations > attempts:
            return False

        udf_list = connection.udf_list({"timeout": 100})

        for udf in udf_list:

            # if the UDF now exists, exit the method
            if udf["name"] == udf_name:
                exists = True
                return True

        # increment and pause
        iterations += 1
        time.sleep(sleep_time)



# -*- coding: utf-8 -*-

import pytest
from aerospike import exception as e
from .as_status_codes import AerospikeStatus

import aerospike


@pytest.mark.usefixtures("connection_config")
class TestSelect(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup class.
        """
        self.test_key = ("test", "demo", 1)

        rec = {
            "a": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")],
            "b": {"key": "asd';q;'1';"},
            "c": 1234,
            "d": "!@#@#$QSDAsd;as",
            "n": None,
            ("a" * 14): "long_bin_14",
        }

        as_connection.put(self.test_key, rec)

        yield

        as_connection.remove(self.test_key)

    def test_select_with_key_and_bins(self):
        """
        Test of selecting a single bin
        """
        bins_to_select = ["a"]

        key, meta, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"a": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")]}

        assert meta is not None

        assert key is not None

    def test_select_with_key_and_duplicate_bins(self):
        """
        Test that passing in duplicate bin names does not cause
        errors
        """
        bins_to_select = ["a", "a", "a"]

        key, meta, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"a": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")]}

    def test_select_with_none_policy(self):

        bins_to_select = ["b"]

        key, meta, bins = self.as_connection.select(self.test_key, bins_to_select, None)

        assert bins == {"b": {"key": "asd';q;'1';"}}
        assert meta is not None
        assert key is not None

    def test_select_with_non_existent_key(self):

        key = ("test", "demo", "non-existent")

        bins_to_select = ["a", "b"]

        with pytest.raises(e.RecordNotFound):
            key, meta, bins = self.as_connection.select(key, bins_to_select)

    @pytest.mark.parametrize("bins", ("a", {}, False, 1, None))
    def test_select_with_key_and_single_bin_to_select_not_a_list(self, bins):

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.select(self.test_key, bins)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_select_with_key_and_multiple_bins_to_select(self):

        bins_to_select = ["c", "d"]

        _, meta, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"c": 1234, "d": "!@#@#$QSDAsd;as"}
        assert meta is not None

    def test_select_with_key_and_multiple_bins_to_select_policy_key_send(self):

        bins_to_select = ["c", "d"]
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_SEND}
        key, meta, bins = self.as_connection.select(self.test_key, bins_to_select, policy)

        # Calculate the key's digest
        key_digest = aerospike.calc_digest(*self.test_key)

        assert bins == {"c": 1234, "d": "!@#@#$QSDAsd;as"}
        assert key == ("test", "demo", 1, key_digest)
        assert meta is not None

    def test_select_with_key_and_multiple_bins_to_select_policy_key_digest(self):

        # key with digest and no primary key
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {
            "a": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")],
            "b": {"key": "asd';q;'1';"},
            "c": 1234,
            "d": "!@#@#$QSDAsd;as",
        }

        self.as_connection.put(key, rec)

        bins_to_select = ["c", "d"]
        policy = {"key": aerospike.POLICY_KEY_DIGEST}
        key, meta, bins = self.as_connection.select(key, bins_to_select, policy)

        assert bins == {"c": 1234, "d": "!@#@#$QSDAsd;as"}
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))
        assert meta is not None

        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        self.as_connection.remove(key)

    def test_select_with_key_with_existent_and_non_existent_bins_to_select(self):

        bins_to_select = ["c", "d", "n", "fake"]

        _, meta, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"c": 1234, "d": "!@#@#$QSDAsd;as", "n": None}
        assert meta is not None

    def test_select_with_key_and_non_existent_bin_in_middle(self):

        bins_to_select = ["c", "fake", "d"]

        key, meta, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"c": 1234, "d": "!@#@#$QSDAsd;as"}
        assert meta is not None

    def test_select_with_key_and_non_existent_bins_to_select(self):
        """
        Test specifying only nonexistent bin names as the argument
        """
        bins_to_select = ["fake1", "fake2"]

        _, _, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {}

    def test_select_with_unicode_bin_value_tuple(self):
        """
        Test for passing a unicode bin name in a tuple will work
        """
        bins_to_select = ("a",)

        _, _, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"a": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")]}

    # Tests for error causing conditions

    def test_select_without_any_parameter(self):

        with pytest.raises(TypeError) as typeError:
            self.as_connection.select()

        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_select_with_key_and_bins_without_connection(self):

        bins_to_select = ["a"]
        config = self.connection_config
        disconnected_client = aerospike.client(config)
        disconnected_client.close()

        with pytest.raises(e.ClusterError) as err_info:
            disconnected_client.select(self.test_key, bins_to_select)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_select_with_invalid_keys(self, invalid_key):
        """
        Verify that different types of invalid keys will
        cause select to fail
        """
        bins_to_select = ["a"]

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.select(invalid_key, bins_to_select)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_select_with_key_and_empty_list_of_bins_to_select(self):
        """
        Test that passing an empty list of bins to select will
        cause an error
        """
        with pytest.raises(e.InvalidRequest) as err_info:
            key, meta, bins = self.as_connection.select(self.test_key, [])

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_select_empty_bin_name_tuple_raises_error(self):

        bins_to_select = ()
        #  This test fails with pytest.raises, but does actually raise
        #  the error
        with pytest.raises(e.InvalidRequest):
            self.as_connection.select(self.test_key, bins_to_select)

    def test_select_invalid_bin_name_tuple_raises_error(self):

        bins_to_select = (1, 2)

        with pytest.raises(e.ParamError):
            self.as_connection.select(self.test_key, bins_to_select)

    def test_select_with_invalid_bin_name_types_raises_error(self):
        bins_to_select = [1, 2]
        with pytest.raises(e.ParamError):
            self.as_connection.select(self.test_key, bins_to_select)

    def test_select_with_unicode_bin_value_list(self):

        bins_to_select = ["a"]

        _, _, bins = self.as_connection.select(self.test_key, bins_to_select)

        assert bins == {"a": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")]}

    @pytest.mark.skip(reason="Behavior is unexpected, but not wrong")
    def test_select_with_very_long_bin_name(self):
        bins_to_select = ["a" * 10000]  # max bin name size is 14
        # internally this only sends the first 14 characters to
        # the select function
        _, _, rec = self.as_connection.select(self.test_key, bins_to_select)
        assert rec["a" * 14] == "long_bin_14"



# -*- coding: utf-8 -*-

import pytest
import json
from contextlib import contextmanager
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@contextmanager
def open_as_connection(config):
    """
    Context manager to let us open aerospike connections with
    specified config
    """
    as_connection = TestBaseClass.get_new_connection(config)

    # Connection is setup, so yield it
    yield as_connection

    # close the connection
    as_connection.close()


# adds cls.connection_config to this class
@pytest.mark.usefixtures("connection_config")
class TestConnect(object):
    def test_version(self):
        """
        Check for aerospike vrsion
        """
        assert aerospike.__version__ is not None

    def test_connect_positive(self):
        """
        Invoke connect() with positive parameters.
        """
        config = self.connection_config.copy()

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()

    def test_connect_positive_with_policy(self):
        """
        Invoke connect() with positive parameters and policy.
        """

        config = self.connection_config.copy()
        config["policies"] = {"read": {"total_timeout": 10000}}

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()

    @pytest.mark.skip(reason="This doesn't actually use multiple hosts")
    def test_connect_positive_with_multiple_hosts(self):
        """
        Invoke connect() with multiple hosts.
        """
        config = self.connection_config.copy()
        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()

    def test_connect_positive_unicode_hosts(self):
        """
        Invoke connect() with unicode hosts.
        """

        #  Convert to and from json, to force unicode
        #  This works in py 2.7 and 3.5
        uni = json.dumps(self.connection_config["hosts"][0])
        hostlist = json.loads(uni)
        config = {"hosts": [tuple(hostlist)]}
        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()

    @pytest.mark.skipif(TestBaseClass.tls_in_use(), reason="no default port for tls")
    def test_connect_hosts_missing_port(self):
        """
        Invoke connect() with missing port in config dict.
        """
        config = {"hosts": [("127.0.0.1")]}

        with open_as_connection(config) as client:
            assert client.is_connected()

    def test_connect_positive_shm_key(self):
        """
        Invoke connect() with shm_key specified
        """
        config = self.connection_config.copy()
        config["shm"] = {"shm_key": 3}

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()
            assert client.shm_key() == 3

    def test_connect_positive_shm_key_default(self):
        """
        Invoke connect() with shm enabled but shm_key not specified
        """
        config = self.connection_config.copy()
        config["shm"] = {"shm_max_nodes": 5}

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()
            assert client.shm_key() is not None

    def test_connect_positive_shm_not_enabled(self):
        """
        Invoke connect() with shm not anabled
        """
        config = self.connection_config.copy()

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()
            assert client.shm_key() is None

    @pytest.mark.skip(reason=("This raises an error," + " but it isn't clear whether it should"))
    def test_connect_positive_cluster_name(self):
        """
        Invoke connect() giving a cluster name
        """
        config = self.connection_config.copy()
        config["cluster_name"] = "test-cluster"

        with pytest.raises(e.ClientError) as err_info:
            self.client = aerospike.client(config).connect()

        assert err_info.value.code == -1

    def test_connect_positive_reconnect(self):
        """
        Client call itself establishes connection.
        """
        config = self.connection_config.copy()

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()
            client.close()
            assert client.is_connected() is False
            if TestBaseClass.user is None and TestBaseClass.password is None:
                client.connect()
            else:
                client.connect(TestBaseClass.user, TestBaseClass.password)
            assert client.is_connected()

    def test_connect_on_connected_client(self):
        """
        Invoke connect on a client that is already connected.
        """
        config = self.connection_config.copy()

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()
            client.connect()
            assert client.is_connected()

    # Tests for invalid Usage

    def test_connect_with_extra_args(self):
        with pytest.raises(TypeError):
            client = aerospike.client(self.connection_config)
            client.close()
            client.connect("username", "password", "extra arg")

    @pytest.mark.parametrize(
        "config, err, err_code, err_msg",
        [
            (1, e.ParamError, -2, "Config must be a dict"),
            ({}, e.ParamError, -2, "Hosts must be a list"),
            ({"": [("127.0.0.1", 3000)]}, e.ParamError, -2, "Hosts must be a list"),
            ({"hosts": [3000]}, e.ParamError, -2, "Invalid host"),
            ({"hosts": [("127.0.0.1", 2000)]}, e.ClientError, -10, "Failed to connect"),
            ({"hosts": [("127.0.0.1", "3000")]}, e.ClientError, -10, "Failed to connect"),
        ],
        ids=[
            "config not dict",
            "config empty dict",
            "config missing hosts key",
            "hosts missing address",
            "hosts port is incorrect",
            "hosts port is string",
        ],
    )
    def test_connect_invalid_configs(self, config, err, err_code, err_msg):
        with pytest.raises(err) as err_info:
            self.client = aerospike.client(config).connect()

        assert err_info.value.code == err_code
        assert err_info.value.msg == err_msg



# -*- coding: utf-8 -*-
import pytest


class TestListBasics(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):

        key = ("test", "map_test", 1)
        rec = {"list": [1, 2, 3, 4, 5]}
        as_connection.put(key, rec)

        def teardown():
            """
            Teardown method.
            """
            key = ("test", "map_test", 1)
            self.as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "method_name",
        (
            "list_append",
            "list_extend",
            "list_insert",
            "list_insert_items",
            "list_pop",
            "list_pop_range",
            "list_remove",
            "list_remove_range",
            "list_clear",
            "list_set",
            "list_get",
            "list_get_range",
            "list_trim",
            "list_size",
        ),
    )
    def test_calling_list_methods_with_no_args(self, method_name):
        method = getattr(self.as_connection, method_name)
        with pytest.raises(TypeError):
            method()



# -*- coding: utf-8 -*-

import pytest

from .test_base_class import TestBaseClass

import aerospike
from aerospike import exception as e

# aerospike.OP_MAP_SET_POLICY
# aerospike.OP_MAP_PUT
# aerospike.OP_MAP_PUT_ITEMS
# aerospike.OP_MAP_INCREMENT
# aerospike.OP_MAP_DECREMENT
# aerospike.OP_MAP_SIZE
# aerospike.OP_MAP_CLEAR
# aerospike.OP_MAP_REMOVE_BY_KEY
# aerospike.OP_MAP_REMOVE_BY_KEY_LIST
# aerospike.OP_MAP_REMOVE_BY_KEY_RANGE
# aerospike.OP_MAP_REMOVE_BY_VALUE
# aerospike.OP_MAP_REMOVE_BY_VALUE_LIST
# aerospike.OP_MAP_REMOVE_BY_VALUE_RANGE
# aerospike.OP_MAP_REMOVE_BY_INDEX
# aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE
# aerospike.OP_MAP_REMOVE_BY_RANK
# aerospike.OP_MAP_REMOVE_BY_RANK_RANGE
# aerospike.OP_MAP_GET_BY_KEY
# aerospike.OP_MAP_GET_BY_KEY_RANGE
# aerospike.OP_MAP_GET_BY_VALUE
# aerospike.OP_MAP_GET_BY_VALUE_RANGE
# aerospike.OP_MAP_GET_BY_INDEX
# aerospike.OP_MAP_GET_BY_INDEX_RANGE
# aerospike.OP_MAP_GET_BY_RANK
# aerospike.OP_MAP_GET_BY_RANK_RANGE


class TestOperate(object):
    def setup_class(cls):
        """
        Setup class.
        """
        cls.client_no_typechecks = TestBaseClass.get_new_connection({"strict_types": False})

    def teardown_class(cls):
        TestOperate.client_no_typechecks.close()

    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        key = ("test", "demo", "test_op_map")
        self.test_map_key = key
        test_map = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}
        self.test_map = test_map
        self.test_map_bin = "test_map"
        key_order_policy = {"map_order": aerospike.MAP_KEY_ORDERED}
        as_connection.map_put_items(key, bin="test_map", items=test_map, map_policy=key_order_policy)
        as_connection.map_put_items(key, bin="test_map2", items=test_map, map_policy=key_order_policy)

        def teardown():
            try:
                as_connection.remove(key)
            except Exception:
                pass
            """
            Teardown Method
            """

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "map_test", 1),
                [
                    {"op": aerospike.OP_MAP_PUT, "bin": "my_map", "key": "age", "val": 97},
                    {"op": aerospike.OP_MAP_INCREMENT, "bin": "my_map", "key": "age", "val": 1},
                    {
                        "op": aerospike.OP_MAP_GET_BY_KEY,
                        "bin": "my_map",
                        "key": "age",
                        "return_type": aerospike.MAP_RETURN_VALUE,
                    },
                ],
                {"my_map": 98},
            ),
            (
                ("test", "map_test", 1),
                [
                    {"op": aerospike.OP_MAP_PUT, "bin": "my_map", "key": "age", "val": 22},
                    {"op": aerospike.OP_MAP_DECREMENT, "bin": "my_map", "key": "age", "val": 1},
                    {
                        "op": aerospike.OP_MAP_GET_BY_KEY,
                        "bin": "my_map",
                        "key": "age",
                        "return_type": aerospike.MAP_RETURN_VALUE,
                    },
                ],
                {"my_map": 21},
            ),
            (
                ("test", "map_test", 1),
                [
                    {
                        "op": aerospike.OP_MAP_PUT_ITEMS,
                        "bin": "my_map",
                        "val": {"name": "bubba", "occupation": "dancer"},
                    },
                    {
                        "op": aerospike.OP_MAP_GET_BY_KEY,
                        "bin": "my_map",
                        "key": "name",
                        "return_type": aerospike.MAP_RETURN_KEY_VALUE,
                    },
                ],
                {"my_map": ["name", "bubba"]},
            ),
        ],
    )
    def test_pos_operate_with_correct_paramters(self, key, llist, expected):
        """
        Invoke operate() with correct parameters
        """
        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == expected
        self.as_connection.remove(key)

    @pytest.mark.skip()
    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "map_test", 1),
                [
                    {"op": aerospike.OP_MAP_PUT, "bin": "my_map", "key": "age", "val": 97},
                    {"op": aerospike.OP_MAP_INCREMENT, "bin": "my_map", "key": "age", "val": 1},
                    {
                        "op": aerospike.OP_MAP_GET_BY_KEY,
                        "bin": "my_map",
                        "key": "age",
                        "return_type": aerospike.MAP_RETURN_VALUE,
                    },
                ],
                [None, None, ("my_map", 98)],
            ),
            (
                ("test", "map_test", 1),
                [
                    {"op": aerospike.OP_MAP_PUT, "bin": "my_map", "key": "age", "val": 22},
                    {"op": aerospike.OP_MAP_DECREMENT, "bin": "my_map", "key": "age", "val": 1},
                    {
                        "op": aerospike.OP_MAP_GET_BY_KEY,
                        "bin": "my_map",
                        "key": "age",
                        "return_type": aerospike.MAP_RETURN_VALUE,
                    },
                ],
                [None, None, ("my_map", 21)],
            ),
            (
                ("test", "map_test", 1),
                [
                    {
                        "op": aerospike.OP_MAP_PUT_ITEMS,
                        "bin": "my_map",
                        "val": {"name": "bubba", "occupation": "dancer"},
                    },
                    {
                        "op": aerospike.OP_MAP_GET_BY_KEY,
                        "bin": "my_map",
                        "key": "name",
                        "return_type": aerospike.MAP_RETURN_KEY_VALUE,
                    },
                ],
                [None, ("my_map", [("name", "bubba")])],
            ),
        ],
    )
    def test_pos_operate_ordered_with_correct_paramters(self, key, llist, expected):
        """
        Invoke operate() with correct parameters
        """
        key, _, bins = self.as_connection.operate_ordered(key, llist)

        assert bins == expected
        self.as_connection.remove(key)

    def test_pos_operate_set_map_policy(self):
        key = ("test", "map_test", 1)
        llist = [
            {"op": aerospike.OP_MAP_SET_POLICY, "bin": "my_map", "map_policy": {"map_sort": aerospike.MAP_KEY_ORDERED}}
        ]
        key, _, _ = self.as_connection.operate(key, llist)
        self.as_connection.remove(key)

    def test_pos_map_clear(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        llist = [{"op": aerospike.OP_MAP_PUT, "bin": binname, "key": "age", "val": 97}]
        key, _, _ = self.as_connection.operate(key, llist)

        llist = [{"op": aerospike.OP_MAP_SIZE, "bin": binname}]
        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == {binname: 1}

        key, _, _ = self.as_connection.operate(key, llist)
        llist = [{"op": aerospike.OP_MAP_CLEAR, "bin": binname}]
        key, _, _ = self.as_connection.operate(key, llist)

        llist = [{"op": aerospike.OP_MAP_SIZE, "bin": binname}]
        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == {binname: 0}
        self.as_connection.remove(key)

    def test_map_remove_by_index_range_correct(self):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE,
                "bin": self.test_map_bin,
                "index": 1,
                "val": 3,
            }
        ]
        _, _, bins = self.as_connection.operate(self.test_map_key, ops)

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == {"a": 1, "e": 5}

    @pytest.mark.xfail(reason="previously worked")
    def test_map_remove_by_index_range_no_index(self):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE,
                "bin": self.test_map_bin,
                "val": 3,
            }
        ]
        with pytest.raises(e.ParamError):
            _, _, bins = self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_put_existing_key(self):
        result_map = self.test_map.copy()
        result_map["a"] = "b"
        ops = [{"op": aerospike.OP_MAP_PUT, "bin": self.test_map_bin, "key": "a", "val": "b"}]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins["test_map"] == result_map

    def test_op_map_put_new_key(self):
        result_map = self.test_map.copy()
        result_map["new"] = "value"
        ops = [{"op": aerospike.OP_MAP_PUT, "bin": self.test_map_bin, "key": "new", "val": "value"}]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("required_key", ("bin", "key", "val"))
    def test_op_map_put_missing_required_keys(self, required_key):
        op = {"op": aerospike.OP_MAP_PUT, "bin": "test_map", "key": "new", "val": "value"}
        del op[required_key]

        ops = [op]
        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_put_items(self):
        result_map = self.test_map.copy()
        result_map["new"] = "value"
        result_map["new2"] = "value2"
        ops = [{"op": aerospike.OP_MAP_PUT_ITEMS, "bin": self.test_map_bin, "val": {"new": "value", "new2": "value2"}}]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("key", ("bin", "val"))
    def test_op_map_put_items_missing_required_entry(self, key):
        op = {"op": aerospike.OP_MAP_PUT_ITEMS, "bin": "test_map", "val": {"new": "value", "new2": "value2"}}
        del op[key]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_increment(self):
        result_map = self.test_map.copy()
        result_map["a"] = result_map["a"] + 2
        ops = [{"op": aerospike.OP_MAP_INCREMENT, "bin": self.test_map_bin, "key": "a", "val": 2}]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("val", ("str", [1, 2, 3], (), {"a": "b"}))
    def test_map_increment_invalid_type(self, val):
        ops = [{"op": aerospike.OP_MAP_INCREMENT, "bin": self.test_map_bin, "key": "a", "val": val}]

        with pytest.raises(Exception):
            self.as_connection.operate(self.test_map_key, ops)

    @pytest.mark.parametrize("key", ("bin", "key", "val"))
    def test_op_map_put_incr_missing_required_entry(self, key):
        op = {"op": aerospike.OP_MAP_INCREMENT, "bin": self.test_map_bin, "key": "a", "val": 1}
        del op[key]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_decrement(self):
        result_map = self.test_map.copy()
        result_map["a"] = result_map["a"] - 2
        ops = [{"op": aerospike.OP_MAP_DECREMENT, "bin": self.test_map_bin, "key": "a", "val": 2}]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("val", ("str", [1, 2, 3], (), {"a": "b"}))
    def test_map_decrement_invalid_type(self, val):
        ops = [{"op": aerospike.OP_MAP_DECREMENT, "bin": self.test_map_bin, "key": "a", "val": val}]

        with pytest.raises(Exception):
            self.as_connection.operate(self.test_map_key, ops)

    @pytest.mark.parametrize("key", ("bin", "key", "val"))
    def test_op_map_put_decr_missing_required_entry(self, key):
        op = {"op": aerospike.OP_MAP_DECREMENT, "bin": self.test_map_bin, "key": "a", "val": 1}
        del op[key]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_size(self):
        op = {"op": aerospike.OP_MAP_SIZE, "bin": self.test_map_bin}
        ops = [op]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins[self.test_map_bin] == len(self.test_map)

    def test_map_size_no_bin(self):
        op = {
            "op": aerospike.OP_MAP_SIZE,
        }
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_size_non_existent_bin(self):
        op = {"op": aerospike.OP_MAP_SIZE, "bin": "fake_bin"}
        ops = [op]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins["fake_bin"] is None

    def test_map_clear(self):
        ops = [{"op": aerospike.OP_MAP_CLEAR, "bin": self.test_map_bin}]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert len(bins[self.test_map_bin]) == len(self.test_map)

        self.as_connection.operate(self.test_map_key, ops)

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert len(bins[self.test_map_bin]) == 0

    def test_map_remove_by_key(self):
        result_map = self.test_map.copy()
        del result_map["c"]
        ops = [{"op": aerospike.OP_MAP_REMOVE_BY_KEY, "bin": self.test_map_bin, "key": "c"}]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert "c" not in bins[self.test_map_bin]

    def test_map_remove_by_key_ret_key(self):
        result_map = self.test_map.copy()
        del result_map["c"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY,
                "bin": "test_map",
                "key": "c",
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins[self.test_map_bin] == "c"

    def test_map_remove_by_key_ret_val(self):
        result_map = self.test_map.copy()
        del result_map["c"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY,
                "bin": "test_map",
                "key": "c",
                "return_type": aerospike.MAP_RETURN_VALUE,
            }
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins["test_map"] == self.test_map["c"]

    def test_map_remove_by_key_ret_key_val(self):
        result_map = self.test_map.copy()
        del result_map["c"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY,
                "bin": self.test_map_bin,
                "key": "c",
                "return_type": aerospike.MAP_RETURN_KEY_VALUE,
            }
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins[self.test_map_bin] == ["c", self.test_map["c"]]

    def test_map_remove_by_key_ret_key_val_test_with_list_read_odd(self):
        self.test_map.copy()
        self.as_connection.put(self.test_map_key, {"cool_list": [1, 2, 3]})
        ops = [
            {"op": aerospike.OPERATOR_READ, "bin": "cool_list"},
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY,
                "bin": self.test_map_bin,
                "key": "c",
                "return_type": aerospike.MAP_RETURN_KEY_VALUE,
            },
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins["cool_list"] == [1, 2, 3]

    def test_map_remove_by_key_ret_key_val_test_with_list_read_even(self):
        self.test_map.copy()
        self.as_connection.put(self.test_map_key, {"cool_list": [1, 2, 3, 4]})
        ops = [
            {"op": aerospike.OPERATOR_READ, "bin": "cool_list"},
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY,
                "bin": self.test_map_bin,
                "key": "c",
                "return_type": aerospike.MAP_RETURN_KEY_VALUE,
            },
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins["cool_list"] == [1, 2, 3, 4]

    @pytest.mark.parametrize("entry", ("bin", "key"))
    def test_map_remove_by_key_missing_required_entries(self, entry):
        op = {
            "op": aerospike.OP_MAP_REMOVE_BY_KEY,
            "bin": self.test_map_bin,
            "key": "c",
            "return_type": aerospike.MAP_RETURN_KEY_VALUE,
        }

        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    @pytest.mark.parametrize("entry", ([], {}, ()))
    def test_map_remove_by_key_invalid_bin(self, entry):
        op = {
            "op": aerospike.OP_MAP_REMOVE_BY_KEY,
            "bin": entry,
            "key": "a",
            "return_type": aerospike.MAP_RETURN_KEY_VALUE,
        }
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_key_list(self):
        result_map = self.test_map
        del result_map["a"]
        del result_map["c"]
        del result_map["d"]

        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY_LIST,
                "bin": self.test_map_bin,
                "val": ["a", "c", "d"],
            }
        ]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)

        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("val", ("a", {"a": "b"}, 2, ("a", "b")))
    def test_map_remove_by_key_list_wrong_val_type(self, val):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY_LIST,
                "bin": self.test_map_bin,
                "val": val,
            }
        ]

        with pytest.raises(Exception):
            self.as_connection.operate(self.test_map_key, ops)

    @pytest.mark.parametrize("val", ("bin", "val"))
    def test_map_remove_by_key_missing_required_entry(self, val):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY_LIST,
                "bin": self.test_map_bin,
                "val": val,
            }
        ]
        del ops[0][val]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_key_range(self):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_KEY_RANGE,
                "bin": self.test_map_bin,
                "key": "b",
                "val": "d",
            }
        ]

        result_map = self.test_map.copy()
        del result_map["b"]
        del result_map["c"]

        self.as_connection.operate(self.test_map_key, ops)
        _, _, bins = self.as_connection.get(self.test_map_key)
        assert result_map == bins[self.test_map_bin]

    @pytest.mark.parametrize("entry", ("bin", "key", "val"))
    def test_map_remove_by_key_range_missing_entry(self, entry):
        op = {"op": aerospike.OP_MAP_REMOVE_BY_KEY_RANGE, "bin": self.test_map_bin, "key": "b", "val": "d"}

        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_value(self):
        result_map = self.test_map.copy()
        del result_map["d"]

        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_VALUE,
                "bin": self.test_map_bin,
                "val": 4,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins[self.test_map_bin] == ["d"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert result_map == bins[self.test_map_bin]

    @pytest.mark.parametrize("entry", ("bin", "val"))
    def test_map_remove_by_value_missing_required_entry(self, entry):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_VALUE,
                "bin": self.test_map_bin,
                "val": 4,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        del ops[0][entry]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_value_list(self):
        result_map = self.test_map.copy()
        del result_map["b"]
        del result_map["d"]
        del result_map["e"]

        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_VALUE_LIST,
                "bin": self.test_map_bin,
                "val": [2, 4, 5],
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins[self.test_map_bin] == ["b", "d", "e"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert result_map == bins[self.test_map_bin]

    @pytest.mark.parametrize("entry", ("bin", "val"))
    def test_map_remove_by_value_list_missing_required_entry(self, entry):
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_VALUE_LIST,
                "bin": self.test_map_bin,
                "val": [2, 4, 5],
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]
        del ops[0][entry]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_value_range(self):
        result_map = self.test_map.copy()
        del result_map["b"]
        del result_map["c"]
        del result_map["d"]

        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_VALUE_RANGE,
                "bin": self.test_map_bin,
                "val": 2,
                "range": 5,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, bins = self.as_connection.operate(self.test_map_key, ops)
        assert bins[self.test_map_bin] == ["b", "c", "d"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert result_map == bins[self.test_map_bin]

    @pytest.mark.parametrize("entry", ("bin", "val", "range"))
    def test_map_remove_by_value_range_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_REMOVE_BY_VALUE_RANGE,
            "bin": self.test_map_bin,
            "val": 2,
            "range": 5,
            "return_type": aerospike.MAP_RETURN_KEY,
        }

        del op[entry]
        ops = [op]
        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_index(self):
        result_map = self.test_map.copy()
        del result_map["b"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_INDEX,
                "bin": self.test_map_bin,
                "index": 1,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == "b"

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("entry", ("bin", "index"))
    def test_map_remove_by_index_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_REMOVE_BY_INDEX,
            "bin": self.test_map_bin,
            "index": 1,
            "return_type": aerospike.MAP_RETURN_KEY,
        }

        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_index_range(self):
        result_map = self.test_map.copy()
        del result_map["b"]
        del result_map["c"]
        del result_map["d"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE,
                "bin": self.test_map_bin,
                "index": 1,
                "val": 3,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["b", "c", "d"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("entry", ("bin", "index", "val"))
    def test_map_remove_by_index_range_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE,
            "bin": self.test_map_bin,
            "index": 1,
            "val": 3,
            "return_type": aerospike.MAP_RETURN_KEY,
        }
        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    @pytest.mark.xfail(reason="This works, but shouldn't")
    def test_map_remove_by_index_range_no_strict_types_val(self):
        result_map = self.test_map.copy()
        del result_map["b"]
        del result_map["c"]
        del result_map["d"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE,
                "bin": self.test_map_bin,
                "index": 1,
                "val": 3,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.client_no_typechecks.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["b", "c", "d"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    def test_map_remove_by_rank(self):
        """
        remove the 3rd item ordered by val
        """
        result_map = self.test_map.copy()
        del result_map["c"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_RANK,
                "bin": self.test_map_bin,
                "index": 2,
            }
        ]

        self.as_connection.operate(self.test_map_key, ops)

        _, _, bins = self.as_connection.get(self.test_map_key)

        assert bins[self.test_map_bin] == result_map

    @pytest.mark.parametrize("entry", ("bin", "index"))
    def test_map_remove_by_rank_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_REMOVE_BY_RANK,
            "bin": self.test_map_bin,
            "index": 2,
        }
        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_map_remove_by_rank_range(self):
        result_map = self.test_map.copy()
        del result_map["c"]
        del result_map["d"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_RANK_RANGE,
                "bin": self.test_map_bin,
                "index": 2,
                "val": 2,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["c", "d"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    def test_map_remove_by_rank_range_val_instead_of_range(self):
        """
        TODO: Should not work
        """
        result_map = self.test_map.copy()
        del result_map["c"]
        del result_map["d"]
        ops = [
            {
                "op": aerospike.OP_MAP_REMOVE_BY_RANK_RANGE,
                "bin": self.test_map_bin,
                "index": 2,
                "val": 2,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["c", "d"]

        _, _, bins = self.as_connection.get(self.test_map_key)
        assert bins[self.test_map_bin] == result_map

    def test_op_map_get_by_key(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_KEY,
                "bin": self.test_map_bin,
                "key": "b",
                "return_type": aerospike.MAP_RETURN_VALUE,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)

        assert res[self.test_map_bin] == 2

    @pytest.mark.parametrize("entry", ("bin", "key"))
    def test_op_map_get_by_key_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_GET_BY_KEY,
            "bin": self.test_map_bin,
            "key": "b",
            "return_type": aerospike.MAP_RETURN_VALUE,
        }
        del op[entry]
        ops = [op]
        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_key_range(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_KEY_RANGE,
                "bin": self.test_map_bin,
                "key": "b",
                "range": "e",
                "return_type": aerospike.MAP_RETURN_VALUE,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)

        assert res[self.test_map_bin] == [2, 3, 4]

    @pytest.mark.parametrize("entry", ("bin", "key", "range"))
    def test_op_map_get_by_key_range_missing_required_entry(self, entry):
        """
        TODO Figure out if this is correct
        """
        op = {
            "op": aerospike.OP_MAP_GET_BY_KEY_RANGE,
            "bin": self.test_map_bin,
            "key": "b",
            "range": "e",
            "return_type": aerospike.MAP_RETURN_VALUE,
        }
        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_value(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_VALUE,
                "bin": self.test_map_bin,
                "val": 2,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)

        assert res[self.test_map_bin] == ["b"]

    @pytest.mark.parametrize("entry", ("bin", "val"))
    def test_op_map_get_by_value_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_GET_BY_VALUE,
            "bin": self.test_map_bin,
            "val": 2,
            "return_type": aerospike.MAP_RETURN_KEY,
        }

        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_value_range(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_VALUE_RANGE,
                "bin": self.test_map_bin,
                "val": 2,
                "range": 5,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["b", "c", "d"]

    @pytest.mark.parametrize("entry", ("bin", "val", "range"))
    def test_op_map_get_by_value_range_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_GET_BY_VALUE_RANGE,
            "bin": self.test_map_bin,
            "val": 2,
            "range": 5,
            "return_type": aerospike.MAP_RETURN_KEY,
        }
        del op[entry]

        ops = [op]
        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_index(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_INDEX,
                "bin": self.test_map_bin,
                "index": 2,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == "c"

    @pytest.mark.parametrize("entry", ("bin", "index"))
    def test_op_map_get_by_index_missing_required_entry(self, entry):
        op = {
            "op": aerospike.OP_MAP_GET_BY_INDEX,
            "bin": self.test_map_bin,
            "index": 2,
            "return_type": aerospike.MAP_RETURN_KEY,
        }
        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_index_range(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_INDEX_RANGE,
                "bin": self.test_map_bin,
                "index": 1,
                "val": 2,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["b", "c"]

    @pytest.mark.parametrize("entry", ("bin", "index", "val"))
    def test_op_map_get_by_index_range_missing_required(self, entry):
        op = {
            "op": aerospike.OP_MAP_GET_BY_INDEX_RANGE,
            "bin": self.test_map_bin,
            "index": 1,
            "val": 2,
            "return_type": aerospike.MAP_RETURN_KEY,
        }
        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_rank(self):
        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_RANK,
                "bin": self.test_map_bin,
                "index": 1,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == "b"

    @pytest.mark.parametrize("entry", ("bin", "index"))
    def test_op_map_get_by_rank_missing_required_entry(self, entry):
        """
        TODO: this shouldn't be needed
        """
        op = {
            "op": aerospike.OP_MAP_GET_BY_RANK,
            "bin": self.test_map_bin,
            "index": 1,
            "return_type": aerospike.MAP_RETURN_KEY,
        }
        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)

    def test_op_map_get_by_rank_range(self):

        ops = [
            {
                "op": aerospike.OP_MAP_GET_BY_RANK_RANGE,
                "bin": self.test_map_bin,
                "index": 1,
                "val": 2,
                "return_type": aerospike.MAP_RETURN_KEY,
            }
        ]

        _, _, res = self.as_connection.operate(self.test_map_key, ops)
        assert res[self.test_map_bin] == ["b", "c"]

    @pytest.mark.parametrize("entry", ("bin", "index", "val"))
    def test_op_map_get_by_rank_range_val_missing_required_values(self, entry):
        op = {
            "op": aerospike.OP_MAP_GET_BY_RANK_RANGE,
            "bin": self.test_map_bin,
            "index": 1,
            "val": 2,
            "return_type": aerospike.MAP_RETURN_KEY,
        }

        del op[entry]
        ops = [op]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(self.test_map_key, ops)



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e
from aerospike_helpers.operations import map_operations as map_ops

import aerospike


def get_map_result_from_operation(client, key, operations, res_bin):
    """
    Map operations return a 1 bin record.
    This operation extracts the dictionary from it.
    """
    _, _, result_bins = client.operate(key, operations)
    return result_bins[res_bin]


def maps_have_same_keys(map1, map2):
    """
    Return True iff the two maps have an identical set of keys
    """
    if len(map1) != len(map2):
        return False

    for key in map1:
        if key not in map2:
            return False
    return True


def maps_have_same_values(map1, map2):
    if len(map1) != len(map2):
        return False

    for key in map1:
        if key not in map2 or map1[key] != map2[key]:
            return False
    return True


def sort_map(client, test_key, test_bin):
    map_policy = {"map_write_mode": aerospike.MAP_CREATE_ONLY, "map_order": aerospike.MAP_KEY_ORDERED}
    operations = [map_ops.map_set_policy(test_bin, map_policy)]
    client.operate(test_key, operations)


class TestNewListOperationsHelpers(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        self.test_map = {"a": 5, "b": 4, "c": 3, "d": 2, "e": 1}

        self.test_key = "test", "demo", "new_map_op"
        self.test_bin = "map"
        self.as_connection.put(self.test_key, {self.test_bin: self.test_map})
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    def test_map_remove_by_key_list(self):
        operations = [
            map_ops.map_remove_by_key_list(
                self.test_bin, ["a", "b", "c"], return_type=aerospike.MAP_RETURN_VALUE, inverted=True
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["e"], self.test_map["d"]])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "d" not in res_map
        assert "e" not in res_map

    def test_map_remove_by_key_range(self):
        operations = [
            map_ops.map_remove_by_key_range(
                self.test_bin, "a", "d", return_type=aerospike.MAP_RETURN_VALUE, inverted=True
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["e"], self.test_map["d"]])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "d" not in res_map
        assert "e" not in res_map

    def test_map_remove_by_value(self):
        operations = [
            map_ops.map_remove_by_value(
                self.test_bin, self.test_map["a"], return_type=aerospike.MAP_RETURN_KEY, inverted=True
            )
        ]
        assert set(get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)) == set(
            ["b", "c", "d", "e"]
        )

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "b" not in res_map
        assert "c" not in res_map
        assert "d" not in res_map
        assert "e" not in res_map

    def test_map_remove_by_value_list(self):
        operations = [
            map_ops.map_remove_by_value_list(
                self.test_bin,
                [self.test_map["a"], self.test_map["b"], self.test_map["c"]],
                return_type=aerospike.MAP_RETURN_KEY,
                inverted=True,
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["d", "e"])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "d" not in res_map
        assert "e" not in res_map

    def test_map_remove_by_value_range(self):
        operations = [
            map_ops.map_remove_by_value_range(self.test_bin, 1, 4, return_type=aerospike.MAP_RETURN_KEY, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["a", "b"])

        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map
        assert "b" not in res_map

    def test_map_remove_by_index_range(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [
            map_ops.map_remove_by_index_range(self.test_bin, 1, 3, return_type=aerospike.MAP_RETURN_KEY, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert ret_vals == ["a", "e"]
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map
        assert "e" not in res_map

    def test_map_remove_by_rank_range(self):
        operations = [
            map_ops.map_remove_by_rank_range(self.test_bin, 1, 3, return_type=aerospike.MAP_RETURN_KEY, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert set(ret_vals) == set(["a", "e"])
        res_map = self.as_connection.get(self.test_key)[2][self.test_bin]
        assert "a" not in res_map
        assert "e" not in res_map

    def test_map_get_by_key_list(self):
        operations = [
            map_ops.map_get_by_key_list(
                self.test_bin, ["a", "b", "c"], return_type=aerospike.MAP_RETURN_VALUE, inverted=True
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["d"], self.test_map["e"]])

    def test_map_get_by_key_range(self):
        operations = [
            map_ops.map_get_by_key_range(self.test_bin, "a", "d", return_type=aerospike.MAP_RETURN_VALUE, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set([self.test_map["d"], self.test_map["e"]])

    def test_map_get_by_value(self):
        operations = [
            map_ops.map_get_by_value(
                self.test_bin, self.test_map["a"], return_type=aerospike.MAP_RETURN_KEY, inverted=True
            )
        ]
        assert set(get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)) == set(
            ["b", "c", "d", "e"]
        )

    def test_map_get_by_value_list(self):
        operations = [
            map_ops.map_get_by_value_list(
                self.test_bin,
                [self.test_map["a"], self.test_map["b"], self.test_map["c"]],
                return_type=aerospike.MAP_RETURN_KEY,
                inverted=True,
            )
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["d", "e"])

    def test_map_get_by_value_range(self):
        operations = [
            map_ops.map_get_by_value_range(self.test_bin, 1, 4, return_type=aerospike.MAP_RETURN_KEY, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)

        assert set(ret_vals) == set(["a", "b"])

    def test_map_get_by_index_range(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [
            map_ops.map_get_by_index_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_KEY, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert set(ret_vals) == set(["a", "d", "e"])

    def test_map_get_by_rank_range(self):
        sort_map(self.as_connection, self.test_key, self.test_bin)
        operations = [
            map_ops.map_get_by_rank_range(self.test_bin, 1, 2, return_type=aerospike.MAP_RETURN_KEY, inverted=True)
        ]
        ret_vals = get_map_result_from_operation(self.as_connection, self.test_key, operations, self.test_bin)
        assert set(ret_vals) == set(["a", "b", "e"])



# -*- coding: utf-8 -*-

import pytest
from .test_data import key_neg

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


@pytest.mark.usefixtures("as_connection")
class TestRemove:
    @pytest.mark.xfail(reason="open bug #client-533")
    def test_pos_remove_with_existing_record(self):
        """
        Invoke remove() when records are present
        """
        key = ("test", "demo", 1)
        retobj = self.as_connection.remove(key)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, _, _) = self.as_connection.get(key)

        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

    @pytest.mark.xfail(reason="open bug #client-533")
    def test_pos_remove_with_policy(self):
        """
        Invoke remove() with policy
        """
        key = ("test", "demo", 1)
        meta = {"gen": 0}
        policy = {"timeout": 1000}

        retobj = self.as_connection.remove(key, meta, policy)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, meta, _) = self.as_connection.get(key)

        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

        key = ("test", "demo", 1)
        rec = {"name": "name%s" % (str(1)), "addr": "name%s" % (str(1)), "age": 1, "no": 1}
        self.as_connection.put(key, rec)

    @pytest.mark.xfail(reason="open bug #client-533")
    def test_pos_remove_with_policy_all(self):
        """
        Invoke remove() with policy
        """
        key = ("test", "demo", 1)
        meta = {"gen": 0}
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }
        retobj = self.as_connection.remove(key, meta, policy)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, meta, _) = self.as_connection.get(key)

        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

        key = ("test", "demo", 1)
        rec = {"name": "name%s" % (str(1)), "addr": "name%s" % (str(1)), "age": 1, "no": 1}
        self.as_connection.put(key, rec)

    @pytest.mark.xfail(reason="open bug #client-533")
    def test_pos_remove_with_policy_key_digest(self):
        """
        Invoke remove() with policy_key_digest
        """

        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        meta = {"gen": 0}
        policy = {"timeout": 1000, "retry": aerospike.POLICY_RETRY_ONCE, "key": aerospike.POLICY_KEY_DIGEST}
        retobj = self.as_connection.put(key, policy)

        assert retobj == 0

        retobj = self.as_connection.remove(key, meta, policy)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, meta, _) = self.as_connection.get(key)

        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

    @pytest.mark.xfail(reason="open bug #client-533")
    def test_pos_remove_with_policy_gen_ignore(self):
        """
        Invoke remove() with policy gen ignore
        """
        key = ("test", "demo", 1)
        meta = {"gen": 0}
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }

        retobj = self.as_connection.remove(key, meta, policy)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, meta, _) = self.as_connection.get(key)

        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

        key = ("test", "demo", 1)
        rec = {"name": "name%s" % (str(1)), "addr": "name%s" % (str(1)), "age": 1, "no": 1}
        self.as_connection.put(key, rec)

    @pytest.mark.xfail(reason="Issue1 : open bug #client-533")
    def test_pos_remove_with_policy_gen_eq_positive(self):
        """
        Invoke remove() with policy gen positive
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_EQ,
        }

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen}
        retobj = self.as_connection.remove(key, meta, policy)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, meta, _) = self.as_connection.get(key)
        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

        key = ("test", "demo", 1)
        rec = {"name": "name%s" % (str(1)), "addr": "name%s" % (str(1)), "age": 1, "no": 1}
        self.as_connection.put(key, rec)

    @pytest.mark.xfail(reason="open bug #client-533")
    def test_pos_remove_with_policy_gen_GT_positive(self):
        """
        Invoke remove() with policy gen GT positive
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_GT,
        }

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"] + 5
        metadata = {"gen": gen}

        retobj = self.as_connection.remove(key, metadata, policy)

        assert retobj == 0

        with pytest.raises(e.RecordNotFound) as exception:
            (key, meta, _) = self.as_connection.get(key)
        (code, msg, _, _) = exception.value
        assert msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"
        assert code == 2

        key = ("test", "demo", 1)
        rec = {"name": "name%s" % (str(1)), "addr": "name%s" % (str(1)), "age": 1, "no": 1}
        self.as_connection.put(key, rec)

    # Negative Tests
    def test_neg_remove_with_policy_gen_eq_not_equal(self, put_data):
        """
        Invoke remove() with policy gen not equal
        """
        key = ("test", "demo", 1)
        record = {"name": "Michal", "age": 34}
        put_data(self.as_connection, key, record)

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"] + 5
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_EQ,
        }

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"] + 5  # Increment Generation by 5

        with pytest.raises(e.RecordGenerationError) as exception:
            self.as_connection.remove(key, {"gen": gen}, policy)
            (code, msg, _, _) = exception.value
            assert msg == "AEROSPIKE_ERR_RECORD_GENERATION"
            assert code == 3

    def test_neg_remove_with_policy_gen_GT_lesser(self, put_data):
        """
        Invoke remove() with policy gen GT lesser
        """
        key = ("test", "demo", "gen_GT_lesser")
        record = {"Company": "Microsoft", "Place": "US"}
        put_data(self.as_connection, key, record)
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_GT,
        }

        (key, meta) = self.as_connection.exists(key)
        lesserGen = meta["gen"] - 1

        with pytest.raises(e.RecordGenerationError) as exception:
            self.as_connection.remove(key, ("gen", lesserGen), policy)
            (code, msg, _, _) = exception.value
            assert msg == "AEROSPIKE_ERR_RECORD_GENERATION"
            assert code == 3

        (key, meta, bins) = self.as_connection.get(key)

        assert key == ("test", "demo", None, bytearray(b"\xdd)Fs9\xa151\x91{\x1c\n\xc0\xac zV\x10Z+"))
        assert meta is not None
        assert bins == record

    def test_neg_remove_with_policy_as_string(self):
        """
        Invoke remove() with policy as string
        """
        key = ("test", "demo", 1)
        meta = {"gen": 0}
        with pytest.raises(e.ParamError) as exception:
            self.as_connection.remove(key, meta, "String_policy")
            (code, msg, _, _) = exception.value
            assert msg == "policy must be a dict"
            assert code == -2

    def test_neg_remove_with_extra_parameter(self):
        """
        Invoke remove() with extra parameter
        """
        key = ("test", "demo", 1)
        meta = {"gen": 0}
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.remove(key, meta, policy, "Extra Param")
        assert "remove() takes at most 3 arguments (4 given)" in str(typeError.value)

    @pytest.mark.parametrize("key, ex_code, ex_msg", key_neg)
    def test_neg_remove_with_incorrect_data(self, key, ex_code, ex_msg):
        """
        Invoke remove() with namespace as None
        """

        with pytest.raises(e.ParamError) as exception:
            self.as_connection.remove(key)
            (code, msg, _, _) = exception.value
            assert msg == ex_msg
            assert code == ex_code

    @pytest.mark.parametrize(
        "key, ex_name, ex_code",
        [
            (("test", None, 1), e.RecordNotFound, 2),
            (("test", "demo", "incorrect_key"), e.RecordNotFound, 2),
            (("test", "demo", "missing_record"), e.RecordNotFound, 2),
        ],
    )
    def test_neg_remove_with_missing_record(self, key, ex_name, ex_code):
        """
        Invoke remove() with set as None
        """
        with pytest.raises(ex_name) as exception:
            self.as_connection.remove(key)
            (code, _, _, _) = exception.value
            assert code == ex_code

    def test_neg_remove_with_correct_parameters_without_connection(self):
        """
        Invoke remove() with correct arguments without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()
        key = ("test", "demo", 1)

        with pytest.raises(e.ClusterError) as exception:
            client1.remove(key)
            (code, _, _, _) = exception.value
            assert code == 11

    def test_neg_remove_with_no_parameters(self):
        """
        Invoke remove() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.remove()

        assert "argument 'key' (pos 1)" in str(typeError.value)



import aerospike
from aerospike_helpers import expressions as exp

# Connect to database
config = {"hosts": [("127.0.0.1", 3000)]}
client = aerospike.client(config).connect()

# Write player records to database
keys = [("test", "demo", i) for i in range(1, 5)]
records = [
            {'user': "Chief"  , 'scores': [6, 12, 4, 21], 'kd': 1.2},
            {'user': "Arbiter", 'scores': [5, 10, 5, 8] , 'kd': 1.0},
            {'user': "Johnson", 'scores': [8, 17, 20, 5], 'kd': 0.9},
            {'user': "Regret" , 'scores': [4, 2, 3, 5]  , 'kd': 0.3}
        ]
for key, record in zip(keys, records):
    client.put(key, record)

# Example #1: Get players with a K/D ratio >= 1.0

kdGreaterThan1 = exp.GE(exp.FloatBin("kd"), 1.0).compile()
policy = {"expressions": kdGreaterThan1}
# For more details on get_many() usage, see the documentation
records = client.get_many(keys, policy)

for record in records:
    print(record[2])
# {'user': 'Chief', 'scores': [6, 12, 4, 21], 'kd': 1.2}
# {'user': 'Arbiter', 'scores': [5, 10, 5, 8], 'kd': 1.0}
# None
# None

# Example #2: Get player with scores higher than 20
# By nesting expressions, we can create complicated filters

# Get top score
getTopScore = exp.ListGetByRank(
                None,
                aerospike.LIST_RETURN_VALUE,
                exp.ResultType.INTEGER,
                -1,
                exp.ListBin("scores")
                )
# ...then compare it
scoreHigherThan20 = exp.GE(getTopScore, 20).compile()
policy = {"expressions": scoreHigherThan20}
records = client.get_many(keys, policy)

for record in records:
    print(record[2])
# {'user': 'Chief', 'scores': [6, 12, 4, 21], 'kd': 1.2}
# None
# {'user': 'Johnson', 'scores': [8, 17, 20, 5], 'kd': 0.9}
# None



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e
from .index_helpers import ensure_dropped_index
from aerospike_helpers import cdt_ctx

list_index = "list_index"
list_rank = "list_rank"
list_value = "list_value"
map_index = "map_index"
map_key = "map_key"
map_rank = "map_rank"
map_value = "map_value"

ctx_ops = {
    list_index: cdt_ctx.cdt_ctx_list_index,
    list_rank: cdt_ctx.cdt_ctx_list_rank,
    list_value: cdt_ctx.cdt_ctx_list_value,
    map_index: cdt_ctx.cdt_ctx_map_index,
    map_key: cdt_ctx.cdt_ctx_map_key,
    map_rank: cdt_ctx.cdt_ctx_map_rank,
    map_value: cdt_ctx.cdt_ctx_map_value,
}


def add_ctx_op(ctx_type, value):
    ctx_func = ctx_ops[ctx_type]
    return ctx_func(value)


ctx_list_index = []
ctx_list_index.append(add_ctx_op(list_index, 0))

ctx_list_rank = []
ctx_list_rank.append(add_ctx_op(list_rank, -1))

ctx_list_value = []
ctx_list_value.append(add_ctx_op(list_value, 3))

ctx_map_index = []
ctx_map_index.append(add_ctx_op(map_index, 0))

ctx_map_key = []
ctx_map_key.append(add_ctx_op(map_key, "sb"))

ctx_map_rank = []
ctx_map_rank.append(add_ctx_op(map_rank, -1))

ctx_map_value = []
ctx_map_value.append(add_ctx_op(map_value, 3))

ctx_empty = []

if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
    if pytest.__version__ < "3.0.0":
        pytest.skip("It only applies to >= 6.1 enterprise edition")
    else:
        pytestmark = pytest.mark.skip


class TestCDTIndexB64(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "addr": "name%s" % (str(i)),
                "numeric_list": [1, 2, 3, 4],
                "string_list": ["a", "b", "c", "d"],
                "geojson_list": [
                    aerospike.GeoJSON({"type": "Point", "coordinates": [-122.096449, 37.421868]}),
                    aerospike.GeoJSON({"type": "Point", "coordinates": [-122.053321, 37.434212]}),
                ],
                "numeric_map": {"a": 1, "b": 2, "c": 3},
                "string_map": {"sa": "a", "sb": "b", "sc": "c"},
                "age": i,
                "no": i,
            }
            as_connection.put(key, rec)
            keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    def test_get_cdtctxb64_with_correct_parameters(self):
        """
        Invoke get_cdtctx_base64() with correct arguments
        """
        policy = {}
        bs_b4_cdt = self.as_connection.get_cdtctx_base64({"ctx": ctx_list_index})

        r = []
        r.append("sindex-create:ns=test;set=demo;indexname=test_string_list_cdt_index")
        r.append(";indextype=%s" % (cdt_ctx.index_type_string(aerospike.INDEX_TYPE_LIST)))
        r.append(";indexdata=string_list,%s" % (cdt_ctx.index_datatype_string(aerospike.INDEX_STRING)))
        r.append(";context=%s" % (bs_b4_cdt))
        req = "".join(r)

        # print("req is ==========={}", req)
        retobj = self.as_connection.info_all(req, policy=None)
        # print("retobj is ==========={}", retobj)

        self.as_connection.index_remove("test", "test_string_list_cdt_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_cdt_index")

        assert retobj != 0

    def test_get_cdtctxb64_with_invalid_parameters(self):
        """
        Invoke get_cdtctx_base64() with invalid arguments
        """
        try:
            self.as_connection.get_cdtctx_base64(ctx_list_index)
        except e.ParamError:
            pass

    def test_get_cdtctxb64_with_invalid_ctx(self):
        """
        Invoke get_cdtctx_base64() with invalid arguments
        """
        try:
            self.as_connection.get_cdtctx_base64({"ctx": ctx_empty})
        except e.ParamError:
            pass



from aerospike_helpers.operations import operations

# Add name, update age, and return attributes
client.put(keyTuple, {'age': 25, 'career': 'delivery boy'})
ops = [
    operations.increment("age", 1000),
    operations.write("name", "J."),
    operations.prepend("name", "Phillip "),
    operations.append("name", " Fry"),
    operations.read("name"),
    operations.read("career"),
    operations.read("age")
]
(key, meta, bins) = client.operate(key, ops)

print(key) # ('test', 'demo', None, bytearray(b'...'))
# The generation should only increment once
# A transaction is *atomic*
print(meta) # {'ttl': 2592000, 'gen': 2}
print(bins) # Will display all bins selected by read operations
# {'name': 'Phillip J. Fry', 'career': 'delivery boy', 'age': 1025}



import aerospike
from aerospike_helpers import cdt_ctx

list_index = "list_index"
list_rank = "list_rank"
list_value = "list_value"
map_index = "map_index"
map_key = "map_key"
map_rank = "map_rank"
map_value = "map_value"

ctx_ops = {
    list_index: cdt_ctx.cdt_ctx_list_index,
    list_rank: cdt_ctx.cdt_ctx_list_rank,
    list_value: cdt_ctx.cdt_ctx_list_value,
    map_index: cdt_ctx.cdt_ctx_map_index,
    map_key: cdt_ctx.cdt_ctx_map_key,
    map_rank: cdt_ctx.cdt_ctx_map_rank,
    map_value: cdt_ctx.cdt_ctx_map_value,
}

def add_ctx_op(ctx_type, value):
    ctx_func = ctx_ops[ctx_type]
    return ctx_func(value)

def get_aerospike():
    endpoints = [
        ('x.y.z.a', 3000)]

    hosts = [(address[0], address[1]) for address in endpoints]

    config = {
        'hosts': hosts,
    }
    return aerospike.client(config)

ctx_list_index = []
ctx_list_index.append(add_ctx_op(list_index, 0))

client = get_aerospike().connect()
bs_b4_cdt = client.get_cdtctx_base64({'ctx':ctx_list_index})
print("base64 encoded _cdt_ctx: ", bs_b4_cdt)
assert bs_b4_cdt == "khAA"
client.close()


# -*- coding: utf-8 -*-
import pytest
import time
from aerospike import exception as e


class TestTruncate(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        self.truncated_keys = []
        self.un_truncated_keys = []
        for i in range(45):
            key = ("test", "truncate", i)
            rec = {"field": i}
            as_connection.put(key, rec)
            self.keys.append(key)
            self.truncated_keys.append(key)

        for i in range(45):
            key = ("test", "un_trunc", i)
            rec = {"removed": i}
            as_connection.put(key, rec)
            self.keys.append(key)
            self.un_truncated_keys.append(key)
        # Give some time for time differences to resolve themselves
        time.sleep(1.1)
        self.truncate_threshold = int(time.time()) * 10**9

        def teardown():
            """
            Teardown Method
            """
            for key in self.keys:
                try:
                    as_connection.remove(key)
                except Exception:
                    pass

        request.addfinalizer(teardown)

    def _assert_truncation_status(self, keys, exists=False):
        # Give the server some time to finish the call
        # If a truncation was expected, give more time
        # to finish it

        attempts = 5
        if not exists:
            while attempts > 0:
                attempts = attempts - 1
                truncated = True
                time.sleep(1)
                for key in keys:
                    _, meta = self.as_connection.exists(key)
                    if meta is not None:
                        truncated = False
                if truncated:
                    break
            assert truncated
        else:
            # Wait for a potential truncation command to come in
            time.sleep(1)
            for key in keys:
                _, meta = self.as_connection.exists(key)
                assert meta is not None

    def test_whole_set_truncation(self):
        self.as_connection.truncate("test", "truncate", 0)

        self._assert_truncation_status(self.truncated_keys, exists=False)

    def test_truncate_does_not_change_entire_namespace(self):
        self.as_connection.truncate("test", "truncate", 0)

        self._assert_truncation_status(self.un_truncated_keys, exists=True)

    def test_truncate_entire_namespace(self):
        """
        Verify that passing None as the set will
        truncate the entire namespace
        """
        self.as_connection.truncate("test", None, 0)
        #
        self._assert_truncation_status(self.truncated_keys, exists=False)
        self._assert_truncation_status(self.un_truncated_keys, exists=False)

    def test_whole_set_truncation_with_policy(self):
        policy = {"timeout": 1000}
        self.as_connection.truncate("test", "truncate", 0, policy)

        self._assert_truncation_status(self.truncated_keys, exists=False)

    def test_whole_set_truncation_with_none_policy(self):
        policy = None
        self.as_connection.truncate("test", "truncate", 0, policy)

        self._assert_truncation_status(self.truncated_keys, exists=False)

    def test_whole_set_truncation_with_created_threshold(self):
        time.sleep(5)
        threshold = int(time.time()) * 10**9
        time.sleep(5)

        for i in range(45, 90):
            key = ("test", "truncate", i)
            rec = {"field": i}
            self.as_connection.put(key, rec)
            self.keys.append(key)

        self.as_connection.truncate("test", "truncate", threshold)

        self._assert_truncation_status(self.truncated_keys[:45], exists=False)

        # Items created after the LUT should still exist
        self._assert_truncation_status(self.truncated_keys[45:], exists=True)

    def test_truncate_with_lut_before_all_records(self):
        before_lut = self.truncate_threshold - 10**11
        self.as_connection.truncate("test", "truncate", before_lut)

        self._assert_truncation_status(self.truncated_keys, exists=True)

    @pytest.mark.parametrize("namespace, test_set", (("test", "truncate"), ("test", "truncate"), ("test", "truncate")))
    def test_whole_set_unicode_truncation(self, namespace, test_set):
        self.as_connection.truncate(namespace, test_set, 0)

        self._assert_truncation_status(self.truncated_keys, exists=False)

    @pytest.mark.parametrize("fake_namespace, fake_set", (("fake_namespace", "truncate"), ("test", "fake_set")))
    def test_truncate_non_existent_containers(self, fake_namespace, fake_set):
        ret_code = self.as_connection.truncate(fake_namespace, fake_set, 0)
        assert ret_code == 0

    @pytest.mark.parametrize("invalid_ns", (1, 0.5, None, False, {}, (), []))
    def test_invalid_ns_argument_to_truncate(self, invalid_ns):
        with pytest.raises(e.ParamError):
            self.as_connection.truncate(invalid_ns, "truncate", 0)

    @pytest.mark.parametrize("invalid_set", (1, 0.5, False, {}, (), []))
    def test_invalid_set_argument_to_truncate(self, invalid_set):
        with pytest.raises(e.ParamError):
            self.as_connection.truncate("test", invalid_set, 0)

    @pytest.mark.parametrize("invalid_nanos", ("zero", 0.5, None, {}, (), []))
    def test_invalid_nanos_argument_to_truncate(self, invalid_nanos):
        with pytest.raises(e.ParamError):
            self.as_connection.truncate("test", "truncate", invalid_nanos)

    @pytest.mark.parametrize("fake_namespace, fake_set", (("", "truncate"), ("test", "")))
    def test_truncate_empty_container_names(self, fake_namespace, fake_set):
        with pytest.raises(e.ClientError):
            self.as_connection.truncate(fake_namespace, fake_set, 0)
        self._assert_truncation_status(self.truncated_keys, exists=True)

    def test_nanos_argument_before_cf_epoch(self):
        with pytest.raises(e.ClientError):
            self.as_connection.truncate("test", "truncate", 1)

    def test_nanos_argument_too_large(self):
        with pytest.raises(e.ParamError):
            self.as_connection.truncate("test", "truncate", 2**64)

    def test_nanos_argument_between_int64_and_uint64(self):
        # This may stop raising a client error in 2264
        # as the value will no longer be in the future then
        with pytest.raises(e.ClientError):
            self.as_connection.truncate("test", "truncate", 2**63 + 1)

    def test_nanos_argument_between_int32_and_uint32(self):
        with pytest.raises(e.ClientError):
            self.as_connection.truncate("test", "truncate", 2**31 + 1)

    def test_nanos_argument_negative(self):
        with pytest.raises(e.ParamError):
            self.as_connection.truncate("test", "truncate", -self.truncate_threshold)

    def test_nanos_argument_in_future(self):
        future_time = self.truncate_threshold + 10**12  # 1000 seconds more
        with pytest.raises(e.ClientError):
            self.as_connection.truncate("test", "truncate", future_time)

    def test_no_nanos_arg(self):
        with pytest.raises(TypeError):
            self.as_connection.truncate("test", "truncate")

    def test_only_set_arg(self):
        with pytest.raises(TypeError):
            self.as_connection.truncate("test")

    def test_truncate_with_no_args(self):
        with pytest.raises(TypeError):
            self.as_connection.truncate()

    def test_whole_set_truncation_with_invalid_policy(self):
        policy = {"timeout": 0.5}
        with pytest.raises(e.ClientError):
            self.as_connection.truncate("test", "truncate", 0, policy)

    @pytest.mark.parametrize("policy", ("timeout", 5, False, [], ()))
    def test_whole_set_truncation_with_invalid_policy_type(self, policy):
        with pytest.raises(e.ClientError):
            self.as_connection.truncate("test", "truncate", 0, policy)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from aerospike_helpers import expressions as exp
from .as_status_codes import AerospikeStatus

import aerospike


class TestScan(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        for i in range(19):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)

        key = ("test", "demo", 122)
        llist = [{"op": aerospike.OPERATOR_APPEND, "bin": bytearray("asd;adk\0kj", "utf-8"), "val": "john"}]
        # Creates a record with the key 122, with one bytearray key.
        self.bytearray_bin = bytearray("asd;adk\0kj", "utf-8")
        as_connection.operate(key, llist)
        self.record_count = 20

        def teardown():
            for i in range(19):
                key = ("test", "demo", i)
                as_connection.remove(key)

            key = ("test", "demo", 122)
            as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_scan_with_existent_ns_and_set(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback)

        assert len(records) == self.record_count

    def test_scan_with_existent_ns_and_none_set(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, None)

        scan_obj.foreach(callback)

        assert len(records) == self.record_count

    def test_scan_with_timeout_policy(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"timeout": 2000})

        assert len(records) == self.record_count

    def test_scan_with_expressions_policy(self):

        records = []

        expr = exp.Not(exp.Eq(exp.StrBin("name"), "name4"))

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"timeout": 2000, "expressions": expr.compile()})
        assert (
            len(records) == self.record_count - 2
        )  # 2 because the last record has no "name" bin and won't be included in the result

    # NOTE: This could fail if node record counts are small and unbalanced across nodes.
    @pytest.mark.xfail(reason="Might fail depending on record count and distribution.")
    def test_scan_with_max_records_policy(self):

        records = []

        max_records = self.record_count // 2

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"max_records": max_records})
        assert len(records) == self.record_count // 2

    def test_scan_with_expressions_policy_no_set(self):

        records = []

        expr = exp.Not(exp.Eq(exp.StrBin("name"), "name4"))

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"expressions": expr.compile()})

        assert (
            len(records) == self.record_count - 2
        )  # 2 because the last record has no "name" bin and won't be included in the result

    def test_scan_with_socket_timeout_policy(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"socket_timeout": 9876})

        assert len(records) == self.record_count

    def test_scan_with_records_per_second_policy(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"records_per_second": 10})
        assert len(records) == self.record_count

    def test_scan_with_callback_returning_false(self):
        """
        Invoke scan() with callback function returns false
        """

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            if len(records) == 10:
                return False
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {"timeout": 1000})
        assert len(records) == 10

    def test_scan_with_unicode_set(self):
        records = []
        st = "demo"

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, st)

        scan_obj.foreach(callback)
        assert len(records) == self.record_count

    def test_scan_with_select_clause(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.select("name")

        scan_obj.foreach(callback)
        # Only 19/20 records contain a bin called 'name'
        assert 19 <= len(records) < self.record_count

    def test_scan_with_results_method(self):

        ns = "test"
        st = "demo"

        scan_obj = self.as_connection.scan(ns, st)

        scan_obj.select("name", "age")

        records = scan_obj.results()
        # Only 19/20 records contain a bin called 'name' or 'age'
        # Depending on ldt support this could be record count -1 or minus 2
        assert 19 <= len(records) < self.record_count

    def test_scan_with_results_method_and_expressions(self):

        ns = "test"
        st = "demo"

        expr = exp.Not(exp.Eq(exp.StrBin("name"), "name4"))

        scan_obj = self.as_connection.scan(ns, st)

        scan_obj.select("name", "age")

        records = scan_obj.results({"expressions": expr.compile()})
        # Only 19/20 records contain a bin called 'name' or 'age'
        # Depending on ldt support this could be record count -2 or minus 3
        assert 18 <= len(records) < self.record_count - 1

    def test_scan_with_options_positive(self):
        """
        Invoke scan() with options positive
        """

        records = []

        options = {"concurrent": True}

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback, {}, options)

        assert len(records) == self.record_count

    @pytest.mark.xfail(reason="Server does not respect percent < 100")
    def test_scan_with_options_percent_partial(self):
        """
        Invoke scan() with options negative
        """

        def callback(input_tuple):
            key, _, _ = input_tuple
            records.append(key)

        ns = "test"
        st = "demo"
        records = []
        options = {"concurrent": True}
        scan_obj = self.as_connection.scan(ns, st)
        scan_obj.foreach(callback, {}, options)
        assert len(records) >= 16 and len(records) < 18

    def test_scan_with_options_nobins_true(self):
        """
        Invoke scan() with nobins
        """
        ns = "test"
        st = "demo"

        records = []

        options = {"nobins": True}

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(ns, st)

        scan_obj.foreach(callback, {}, options)

        assert len(records) == self.record_count
        for record in records:
            assert record == {}  # The bins portion should be an empty dict

    def test_scan_with_options_nobins_false(self):
        """
        Invoke scan() with nobins
        """
        ns = "test"
        st = "demo"

        records = []

        options = {"nobins": False}

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(ns, st)

        scan_obj.foreach(callback, {}, options)

        assert len(records) == self.record_count

        for record in records:
            assert record != {}

    def test_scan_with_multiple_foreach_on_same_scan_object(self):
        """
        Invoke multiple foreach on same scan object.
        """
        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.foreach(callback)

        assert len(records) == self.record_count

        records = []
        scan_obj.foreach(callback)

        assert len(records) == self.record_count

    def test_scan_with_multiple_results_call_on_same_scan_object(self):

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.select("name", "age")

        records = scan_obj.results()
        assert 19 <= len(records) < self.record_count

        records = []
        records = scan_obj.results()
        assert 19 <= len(records) < self.record_count

    def test_scan_with_select_binnames_bytearray(self):

        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        scan_obj.select(self.bytearray_bin)

        scan_obj.foreach(callback)

        assert len(records) == 1

    def test_scan_without_any_parameter(self):

        with pytest.raises(e.ParamError):
            self.as_connection.scan()
            assert True

    def test_scan_with_non_existent_ns_and_set(self):

        ns = "namespace"
        st = "set"

        records = []
        scan_obj = self.as_connection.scan(ns, st)

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(callback)
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_with_none_ns_and_set(self):

        ns = None
        st = None

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.scan(ns, st)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_scan_with_select_bin_integer(self):
        """
        Invoke scan() with select bin is of type integer.
        """
        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        with pytest.raises(e.ParamError) as err_info:
            scan_obj.select(22, "name")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_scan_with_callback_contains_error(self):
        records = []

        def callback(input_tuple):
            _, _, bins = input_tuple
            raise Exception("callback error")
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(callback, {"timeout": 1000})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_with_callback_non_callable(self):

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(5)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_with_callback_wrong_number_of_args(self):
        def callback():
            pass

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(callback)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_with_invalid_expressions_policy(self):

        records = []

        expr = exp.Not(exp.Eq(exp.StrBin("name"), 2))

        def callback(input_tuple):
            _, _, bins = input_tuple
            records.append(bins)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        with pytest.raises(e.InvalidRequest):
            scan_obj.foreach(callback, {"timeout": 2000, "expressions": expr.compile()})



from aerospike_helpers.batch import records as br
from aerospike_helpers.operations import operations as op

# Keys
# Only insert two records with the first and second key
keyTuples = [
    ('test', 'demo', 'Robert'),
    ('test', 'demo', 'Daniel'),
    ('test', 'demo', 'Patrick'),
]
client.put(keyTuples[0], {'id': 100, 'balance': 400})
client.put(keyTuples[1], {'id': 101, 'balance': 200})
client.put(keyTuples[2], {'id': 102, 'balance': 300})

# Apply different operations to different keys
batchRecords = br.BatchRecords(
    [
        # Remove Robert from system
        br.Remove(
            key = keyTuples[0],
        ),
        # Modify Daniel's ID and balance 
        br.Write(
            key = keyTuples[1],
            ops = [
                op.write("id", 200),
                op.write("balance", 100),
                op.read("id"),
            ],
        ),
        # Read Patrick's ID
        br.Read(
            key = keyTuples[2],
            ops=[
                op.read("id")
            ],
            policy=None
        ),
    ]
)

client.batch_write(batchRecords)

# batch_write modifies its BatchRecords argument.
# Results for each BatchRecord will be set in the result, record, and in_doubt fields.
for batchRecord in batchRecords.batch_records:
    print(batchRecord.result)
    print(batchRecord.record)
# Note how written bins return None if their values aren't read
# And removed records have an empty bins dictionary
# 0
# (('test', 'demo', 'Robert', bytearray(b'...')), {'ttl': 4294967295, 'gen': 0}, {})
# 0
# (('test', 'demo', 'Daniel', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'id': 200, 'balance': None})
# 0
# (('test', 'demo', 'Patrick', bytearray(b'...')), {'ttl': 2592000, 'gen': 1}, {'id': 102})



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Helper functions to create HyperLogLog operation dictionary arguments for
the :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered` methods of the aerospike client.
HyperLogLog bins and operations allow for your application to form fast, reasonable approximations
of members in the union or intersection between multiple HyperLogLog bins.
HyperLogLog’s estimates are a balance between complete accuracy and efficient savings
in space and speed in dealing with extremely large datasets.

    .. note:: HyperLogLog operations require server version >= 4.9.0

    .. seealso:: `HyperLogLog (Data Type) more info. <https://www.aerospike.com/docs/guide/hyperloglog.html>`_.

Example::

    import aerospike
    from aerospike_helpers.operations import hll_operations as hll_ops
    from aerospike_helpers.operations import operations

    NUM_INDEX_BITS = 12
    NUM_MH_BITS = 24

    # Configure the client.
    config = {"hosts": [("127.0.0.1", 3000)]}
    # Create a client and connect it to the cluster.
    client = aerospike.client(config).connect()

    # Create customer keys
    TEST_NS = "test"
    TEST_SET = "demo"
    customerNames = ["Amy", "Farnsworth", "Scruffy"]
    keys = []
    for customer in customerNames:
        keys.append((TEST_NS, TEST_SET, customer))

    itemsViewedPerCustomer = [
        # [item1, item2, ... item500]
        list("item%s" % str(i) for i in range(0, 500)), # Amy
        list("item%s" % str(i) for i in range(0, 750)), # Farnsworth
        list("item%s" % str(i) for i in range(250, 1000)), # Scruffy
    ]

    for key, itemsViewed in zip(keys, itemsViewedPerCustomer):
        customerName = key[2]
        ops = [
            operations.write("name", customerName),
            hll_ops.hll_add("viewed", itemsViewed, NUM_INDEX_BITS, NUM_MH_BITS),
        ]
        client.operate(key, ops)

    # Find out how many items viewed Amy, Farnsworth, and Scruffy have in common.
    farnsworthRecord = client.get(keys[1])
    scruffyRecord = client.get(keys[2])
    farnsworthViewedItems = farnsworthRecord[2]["viewed"]
    scruffyViewedItems = scruffyRecord[2]["viewed"]
    viewed = [farnsworthViewedItems, scruffyViewedItems]
    ops = [
        hll_ops.hll_get_intersect_count("viewed", viewed)
    ]
    # Pass in Amy's key
    _, _, res = client.operate(keys[0], ops)
    print("Estimated items viewed intersection:", res["viewed"])
    # Estimated items viewed intersection: 251
    # Actual intersection: 250

    # Find out how many unique products Amy, Farnsworth, and Scruffy have viewed.
    ops = [hll_ops.hll_get_union_count("viewed", viewed)]
    _, _, res = client.operate(keys[0], ops)

    print("Estimated items viewed union:", res["viewed"])
    # Estimated items viewed union: 1010
    # Actual union: 1000

    # Find the similarity of Amy, Farnsworth, and Scruffy's product views.
    ops = [hll_ops.hll_get_similarity("viewed", viewed)]
    _, _, res = client.operate(keys[0], ops)

    print("Estimated items viewed similarity: %f%%" % (res["viewed"] * 100))
    # Estimated items viewed similarity: 24.888393%
    # Actual similarity: 25%

"""

import aerospike


OP_KEY = "op"
BIN_KEY = "bin"
HLL_POLICY_KEY = "hll_policy"
INDEX_BIT_COUNT_KEY = "index_bit_count"
MH_BIT_COUNT_KEY = "mh_bit_count"
VALUE_LIST_KEY = "value_list"


def hll_add(bin_name: str, values, index_bit_count=None, mh_bit_count=None, policy=None):
    """Creates a hll_add operation.

    Server will add the values to the hll bin.
    If the HLL bin does not exist, it will be created with index_bit_count and/or mh_bit_count if they have been
    supplied.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        values: The values to be added to the HLL set.
        index_bit_count: An optional number of index bits. Must be between 4 and 16 inclusive.
        mh_bit_count: An optional number of min hash bits. Must be between 4 and 58 inclusive.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_ADD,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: values,
        INDEX_BIT_COUNT_KEY: -1 if index_bit_count is None else index_bit_count,
        MH_BIT_COUNT_KEY: -1 if mh_bit_count is None else mh_bit_count,
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict


def hll_describe(bin_name):
    """Creates a hll_describe operation.

    Server returns index and minhash bit counts used to create HLL bin in a list of integers.
    The list size is 2.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_DESCRIBE,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_fold(bin_name: str, index_bit_count):
    """Creates a hll_fold operation.

    Servers folds index_bit_count to the specified value.
    This can only be applied when minhash bit count on the HLL bin is 0.
    Server does not return a value.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index_bit_count: number of index bits. Must be between 4 and 16 inclusive.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_FOLD, BIN_KEY: bin_name, INDEX_BIT_COUNT_KEY: index_bit_count}

    return op_dict


def hll_get_count(bin_name):
    """Creates a hll_get_count operation.

    Server returns estimated count of elements in the HLL bin.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_COUNT,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_get_intersect_count(bin_name: str, hll_list):
    """Creates a hll_get_intersect_count operation.

    Server returns estimate of elements that would be contained by the intersection of these HLL objects.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be intersected.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_INTERSECT_COUNT, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_get_similarity(bin_name: str, hll_list):
    """Creates a hll_get_similarity operation.

    Server returns estimated similarity of the HLL objects.
    Server returns a float.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs used for similarity estimation.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_SIMILARITY, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_get_union(bin_name: str, hll_list):
    """Creates a hll_get_union operation.

    Server returns an HLL object that is the union of all specified HLL objects
    in hll_list with the HLL bin.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be unioned.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_UNION, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_get_union_count(bin_name: str, hll_list):
    """Creates a hll_get_union_count operation.

    Server returns the estimated count of elements that would be contained by the union of all specified HLL objects
    in the list with the HLL bin.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be unioned.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_UNION_COUNT, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_init(bin_name: str, index_bit_count=None, mh_bit_count=None, policy=None):
    """Creates a hll_init operation.

    Server creates a new HLL or resets an existing HLL.
    If index_bit_count and mh_bit_count are None, an existing HLL bin will be reset but retain its configuration.
    If 1 of index_bit_count or mh_bit_count are set,
    an existing HLL bin will set that config and retain its current value for the unset config.
    If the HLL bin does not exist, index_bit_count is required to create it, mh_bit_count is optional.
    Server does not return a value.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index_bit_count: An optional number of index bits. Must be between 4 and 16 inclusive.
        mh_bit_count: An optional number of min hash bits. Must be between 4 and 58 inclusive.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_INIT,
        BIN_KEY: bin_name,
        INDEX_BIT_COUNT_KEY: -1 if index_bit_count is None else index_bit_count,
        MH_BIT_COUNT_KEY: -1 if mh_bit_count is None else mh_bit_count,
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict


def hll_refresh_count(bin_name: str):
    """Creates a hll_refresh_count operation.

    Server updates the cached count if it is stale.
    Server returns the count.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.
    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_REFRESH_COUNT,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_set_union(bin_name: str, hll_list, policy=None):
    """Creates a hll_set_union operation.

    Server sets the union of all specified HLL objects with the HLL bin.
    Server returns nothing.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs who's union will be set.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_SET_UNION, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict
